/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2022 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

MF_BUILTIN_MODULE
MF_COND(WITH_MFMOD)

#include <mflib/status.h>
#include "global.h"
#include "mfmod.h"
#include "msg.h"
#include <dlfcn.h>
#include <mailutils/wordsplit.h>

static size_t nmfmods = MAX_MFMODS;
char *mfmod_path = MFMODDIR;
static struct mu_cfg_param mfmod_cfg_param[] = {
	{ "max-mfmods", mu_c_size, &nmfmods, 0, NULL,
	  N_("Maximum number of dynamically loaded modules.") },
	{ "mfmod-path", mu_c_string, &mfmod_path, 0, NULL,
	  N_("Search path for dynamically loaded modules.") },
	{ NULL }
};

struct modinfo {
	void *handle;
	char *libname;
	char *symname;
	MFMOD_PARAM retval;
};

static void
mfmod_retval_free(struct modinfo *mod)
{
	if (mod->symname) {
		switch (mod->retval.type) {
		case mfmod_string:
			free(mod->retval.string);
			break;
		case mfmod_number:
		case mfmod_message: /*FIXME*/
			break;
		}
		memset(&mod->retval, 0, sizeof(mod->retval));
		free(mod->symname);
		mod->symname = NULL;
	}
}

static inline int
mfmod_is_used(struct modinfo *dltab, int i)
{
	return i >= 0 && i < nmfmods && dltab[i].handle != NULL;
}


static void *
alloc_mfmods(void)
{
	return mu_calloc(nmfmods, sizeof(struct modinfo));
}

static struct modinfo *dltab;

#if 0
// See FIXME below

static void
mfmod_close(struct modinfo *mod)
{
	mfmod_retval_free(mod);
	dlclose(mod->handle);
	free(mod->libname);
	memset(mod, 0, sizeof(*mod));
}

static void
destroy_mfmods(void)
{
	if (dltab) {
		size_t i;
		for (i = 0; i < nmfmods; i++) {
			if (mfmod_is_used(dltab, i))
				mfmod_close(&dltab[i]);
		}
		free(dltab);
	}
}
#endif

static struct modinfo *
get_dltab(void)
{
	if (!dltab) {
		dltab = alloc_mfmods();
		/*
		 * FIXME: Used to have atexit(destroy_mfmods) here.
		 * However, shutdown handlers, that are currently run
		 * via mu_onexit (i.e. atexit) as well get executed after
		 * destroy_mfmod, which causes trouble.
		 * I guess it's OK not to close dlhandles, since we're
		 * going to exit anyway.
		 */
	}
	return dltab;
}

static int
find_free_slot(struct modinfo *dltab)
{
	int i;
	for (i = 0; i < nmfmods; i++)
		if (!mfmod_is_used(dltab, i))
			return i;
	return -1;
}

static char *
mfmod_find(char const *name)
{
	char *retval = NULL;

	if (name[0] == '/')
		retval = mu_strdup(name);
	else {
		struct mu_wordsplit ws;
		size_t i;
	
		ws.ws_delim = ":";
		if (mu_wordsplit(mfmod_path, &ws, MU_WRDSF_DELIM|WRDSF_NOCMD|WRDSF_NOVAR)) {
			mu_error("can't split mfmod_path: %s", mu_wordsplit_strerror(&ws));
			mu_wordsplit_free(&ws);
			return NULL;
		}
		
		for (i = 0; i < ws.ws_wordc; i++) {
			char *libname = mu_make_file_name(ws.ws_wordv[0], name);
			if (access(libname, X_OK) == 0) {
				retval = libname;
				break;
			}
			free(libname);
		}
		mu_wordsplit_free(&ws);
	}
	return retval;
}

MF_DEFUN(dlopen, NUMBER, STRING libname)
{
	struct modinfo *dltab = get_dltab();
	long n = find_free_slot(dltab);
	void *dlh;
	char *filename;
	
	MF_ASSERT(n != -1,
		  mfe_failure, _("no more library slots available"));
	filename = mfmod_find(libname);
	MF_ASSERT(filename != NULL,
		  mfe_failure,
		  _("module file %s not found in mfmod_path"), libname);
	MF_DCL_CLEANUP(CLEANUP_ALWAYS, filename, free);
	dlh = dlopen(filename, RTLD_LAZY | RTLD_LOCAL);
	MF_ASSERT(dlh != NULL, mfe_failure,
		  _("can't load %s: %s"), libname, dlerror());
	dltab[n].handle = dlh;
	dltab[n].libname = mu_strdup(libname);
	MF_RETURN(n);
}
END

static void
_cleanup_vparam(void *ptr)
{
	free(ptr);
}

static void
_cleanup_retval(void *ptr)
{
	struct modinfo *mod = ptr;
	free(mod->retval.string);
	mod->retval.string = NULL;
}

MF_DEFUN_VARARGS_NO_PROM(dlcall, VOID, NUMBER dlh, STRING symname, STRING *types)
{
	long i;
	MFMOD_FUNC func;
	struct modinfo *dltab = get_dltab();
	struct modinfo *mod;
	size_t nparam;
	MFMOD_PARAM *vparam;
	int rc;
	
	MF_ASSERT(mfmod_is_used(dltab, dlh),
		  mfe_failure,
		  _("invalid dynamic library handle"));
	mod = &dltab[dlh];

	mfmod_retval_free(mod);
	dlerror(); /* Clear error state */
	func = dlsym(mod->handle, symname);
	if (!func) {
		char *err = dlerror();
		if (err)
			MF_THROW(mfe_failure, _("can't find symbol %s in %s: %s"),
				 symname, mod->libname, err);
		else
			MF_THROW(mfe_failure, _("%s in module %s has NULL value"),
				 symname, mod->libname);
	}

	nparam = MF_VA_COUNT();
	MF_ASSERT(nparam == strlen(types),
		  mfe_failure,
		  _("number of arguments doesn't satisfy type string"));

	vparam = mu_calloc(nparam, sizeof(vparam[0]));
	MF_DCL_CLEANUP(CLEANUP_ALWAYS, vparam, _cleanup_vparam);

	MF_VA_START();
	for (i = 0; i < nparam; i++) {
		switch (types[i]) {
		case 'n':
		case 'i':
		case 'd':
			vparam[i].type = mfmod_number;
			MF_VA_ARG(i, NUMBER, vparam[i].number);
			break;

		case 's':
			vparam[i].type = mfmod_string;
			MF_VA_ARG(i, STRING, vparam[i].string);
			break;

		case 'm': {
			long n;
			mu_message_t msg;
			
			MF_VA_ARG(i, NUMBER, n);
			msg = bi_message_from_descr(env, n);
			vparam[i].type = mfmod_message;
			vparam[i].message = msg;
			break;
		}
				
		default:
			MF_THROW(mfe_failure,
				 "unrecognized data type in %s at position: %ld",
				 types, i);
		}
	}
	MF_VA_END();

	rc = func(nparam, vparam, &mod->retval);
	if (rc < 0) {
		MF_THROW(mfe_failure,
			 _("call to %s in module %s failed"), symname,
			 mod->libname);
	} else if (rc) {
		if (mod->retval.type == mfmod_string && mod->retval.string) {
			MF_DCL_CLEANUP(CLEANUP_THROW, mod, _cleanup_retval);
			MF_THROW(rc,
				 _("call to %s in module %s failed: %s"),
				 symname, mod->libname, mod->retval.string);
		} else {
			MF_THROW(rc,
				 _("call to %s in module %s failed"), symname,
				 mod->libname);
		}
	}
	mod->symname = mu_strdup(symname);
}
END

static void
dlreturn(eval_environ_t env, int dlh, int type)
{
	struct modinfo *dltab = get_dltab();

	MF_ASSERT(mfmod_is_used(dltab, dlh),
		  mfe_failure,
		  _("invalid dynamic library handle"));
	MF_ASSERT(dltab[dlh].symname != NULL,
		  mfe_failure,
		  _("no return value from the recent call in %s"),
		  dltab[dlh].libname);
	MF_ASSERT(dltab[dlh].retval.type == type,
		  mfe_failure,
		  _("type mismatch in return from %s, library %s"),
		  dltab[dlh].symname, dltab[dlh].libname);
	switch (dltab[dlh].retval.type) {
	case mfmod_number:
		push(env, (STKVAL)(mft_number)dltab[dlh].retval.number);
		break;

	case mfmod_string:
		pushs(env, dltab[dlh].retval.string);
		break;

	case mfmod_message: {
		int rc;
		
		rc = bi_message_register(env, NULL, dltab[dlh].retval.message, MF_MSG_STANDALONE);
		if (rc < 0) {
			mu_message_destroy(&dltab[dlh].retval.message, mu_message_get_owner(dltab[dlh].retval.message));
			MF_THROW(mfe_failure,
				 _("no more message slots available"));
		}
		push(env, (STKVAL)(mft_number)rc);
	}
	}
	mfmod_retval_free(&dltab[dlh]);
}

MF_DEFUN(dlresult_string, STRING, NUMBER dlh)
{
	dlreturn(env, dlh, mfmod_string);
}
END

MF_DEFUN(dlresult_number, NUMBER, NUMBER dlh)
{
	dlreturn(env, dlh, mfmod_number);
}
END

MF_DEFUN(dlresult_message, NUMBER, NUMBER dlh)
{
	dlreturn(env, dlh, mfmod_message);
}
END

MF_INIT([<
	 mf_add_runtime_params(mfmod_cfg_param);
	 >])
